# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pydantic import ConfigDict, model_validator
from typing_extensions import Self

from pixano.utils import issubclass_strict

from ...types import BaseType
from ...types.registry import _register_type_internal
from ...types.schema_reference import EntityRef, ItemRef, SourceRef, ViewRef
from ..registry import _register_schema_internal
from . import Annotation


@_register_type_internal
class BaseIntrinsics(BaseType):
    """BaseIntrinsics (TODO: description?).

    Attributes:
        cx_offset_px: cx_offset_px
        cy_offset_px: cy_offset_px
        img_height_px: img_height_px
        img_width_px: img_width_px
    """

    model_config = ConfigDict(validate_assignment=True)

    cx_offset_px: float
    cy_offset_px: float
    img_height_px: int
    img_width_px: int

    @model_validator(mode="after")
    def _validate_fields(self) -> Self:
        if self.img_height_px < 0:
            raise ValueError("img_height_px must be positive")
        elif self.img_width_px < 0:
            raise ValueError("img_width_px must be positive")
        elif self.cx_offset_px < 0:
            raise ValueError("cx_offset_px must be positive")
        elif self.cy_offset_px < 0:
            raise ValueError("cy_offset_px must be positive")
        return self


@_register_type_internal
class Intrinsics(BaseType):
    """Intrinsics (TODO: description?).

    Attributes:
        c1: c1.
        c2: c2.
        c3: c3.
        c4: c4.
        pixel_aspect_ratio: pixel_aspect_ratio.
    """

    model_config = ConfigDict(validate_assignment=True)

    c1: float
    c2: float
    c3: float
    c4: float
    pixel_aspect_ratio: float

    @model_validator(mode="after")
    def _validate_fields(self):
        if self.c1 < 0:
            raise ValueError("c1 must be positive")
        elif self.c2 < 0:
            raise ValueError("c2 must be positive")
        elif self.c3 < 0:
            raise ValueError("c3 must be positive")
        elif self.c4 < 0:
            raise ValueError("c4 must be positive")
        return self


@_register_type_internal
class Extrinsics(BaseType):
    """Extrinsics (TODO: description?).

    Attributes:
        pos_x_m: pos_x_m.
        pos_y_m: pos_y_m.
        pos_z_m: pos_z_m.
        rot_x_deg: rot_x_deg.
        rot_z1_deg: rot_z1_deg.
        rot_z2_deg: rot_z2_deg.
    """

    model_config = ConfigDict(validate_assignment=True)

    pos_x_m: float
    pos_y_m: float
    pos_z_m: float
    rot_x_deg: float
    rot_z1_deg: float
    rot_z2_deg: float

    @model_validator(mode="after")
    def _validate_fields(self) -> Self:
        if self.pos_x_m < 0:
            raise ValueError("pos_x_m must be positive")
        elif self.pos_y_m < 0:
            raise ValueError("pos_y_m must be positive")
        elif self.pos_z_m < 0:
            raise ValueError("pos_z_m must be positive")
        elif self.rot_x_deg < 0:
            raise ValueError("rot_x_deg must be positive")
        elif self.rot_z1_deg < 0:
            raise ValueError("rot_z1_deg must be positive")
        elif self.rot_z2_deg < 0:
            raise ValueError("rot_z2_deg must be positive")
        return self


@_register_schema_internal
class CamCalibration(Annotation):
    """Camera calibration.

    Attributes:
        type: Type of camera.
        base_intrinsics: Base intrinsics values.
        extrinsics: Extrinsics values.
        intrinsics: Intrinsics values.
    """

    type: str
    base_intrinsics: BaseIntrinsics
    extrinsics: Extrinsics
    intrinsics: Intrinsics

    @classmethod
    def none(cls) -> Self:
        """Utility function to get a `None` equivalent.
        Should be removed as soon as Lance manages `None` value.

        Returns:
            "None" `CamCalibration`.
        """
        return cls(
            id="",
            item_ref=ItemRef.none(),
            view_ref=ViewRef.none(),
            entity_ref=EntityRef.none(),
            type="",
            base_intrinsics=BaseIntrinsics(
                cx_offset_px=0.0,
                cy_offset_px=0.0,
                img_height_px=0,
                img_width_px=0,
            ),
            extrinsics=Extrinsics(
                pos_x_m=0.0,
                pos_y_m=0.0,
                pos_z_m=0.0,
                rot_x_deg=0.0,
                rot_z1_deg=0.0,
                rot_z2_deg=0.0,
            ),
            intrinsics=Intrinsics(
                c1=0.0,
                c2=0.0,
                c3=0.0,
                c4=0.0,
                pixel_aspect_ratio=0.0,
            ),
        )


def is_cam_calibration(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `CamCalibration` or subclass of `CamCalibration`."""
    return issubclass_strict(cls, CamCalibration, strict)


def create_cam_calibration(
    type: str,
    base_intrinsics: BaseIntrinsics | None = None,
    extrinsics: Extrinsics | None = None,
    intrinsics: Intrinsics | None = None,
    cx_offset_px: float | None = None,
    cy_offset_px: float | None = None,
    img_height_px: int | None = None,
    img_width_px: int | None = None,
    pos_x_m: float | None = None,
    pos_y_m: float | None = None,
    pos_z_m: float | None = None,
    rot_x_deg: float | None = None,
    rot_z1_deg: float | None = None,
    rot_z2_deg: float | None = None,
    c1: float | None = None,
    c2: float | None = None,
    c3: float | None = None,
    c4: float | None = None,
    pixel_aspect_ratio: float | None = None,
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    view_ref: ViewRef = ViewRef.none(),
    entity_ref: EntityRef = EntityRef.none(),
    source_ref: SourceRef = SourceRef.none(),
    validate: bool = True,
) -> CamCalibration:
    """Create a `CamCalibration` instance.

    Args:
        type: The type of camera.
        base_intrinsics: The base intrinsics.
        extrinsics: The extrinsics.
        intrinsics: The intrinsics.
        cx_offset_px: cx_offset_px.
        cy_offset_px: cy_offset_px.
        img_height_px: img_height_px.
        img_width_px: img_width_px.
        pos_x_m: pos_x_m.
        pos_y_m: pos_y_m.
        pos_z_m: pos_z_m.
        rot_x_deg: rot_x_deg.
        rot_z1_deg: rot_z1_deg.
        rot_z2_deg: rot_z2_deg.
        c1: c1.
        c2: c2.
        c3: c3.
        c4: c4.
        pixel_aspect_ratio: pixel_aspect_ratio.
        id: `CamCalibration` ID.
        item_ref: Item reference.
        view_ref: View reference.
        entity_ref: Entity reference.
        source_ref: Source reference.
        validate: Set to False to skip pydantic validation.

    Returns:
        The created `CamCalibration` instance.
    """
    none_obj_conditions = [
        base_intrinsics is None,
        extrinsics is None,
        intrinsics is None,
    ]
    not_none_obj_conditions = [
        base_intrinsics is not None,
        extrinsics is not None,
        intrinsics is not None,
    ]

    none_field_conditions = [
        cx_offset_px is None,
        cy_offset_px is None,
        img_height_px is None,
        img_width_px is None,
        pos_x_m is None,
        pos_y_m is None,
        pos_z_m is None,
        rot_x_deg is None,
        rot_z1_deg is None,
        rot_z2_deg is None,
        c1 is None,
        c2 is None,
        c3 is None,
        c4 is None,
        pixel_aspect_ratio is None,
    ]

    not_none_field_conditions = [
        cx_offset_px is not None,
        cy_offset_px is not None,
        img_height_px is not None,
        img_width_px is not None,
        pos_x_m is not None,
        pos_y_m is not None,
        pos_z_m is not None,
        rot_x_deg is not None,
        rot_z1_deg is not None,
        rot_z2_deg is not None,
        c1 is not None,
        c2 is not None,
        c3 is not None,
        c4 is not None,
        pixel_aspect_ratio is not None,
    ]

    if not all(none_obj_conditions) and not all(not_none_obj_conditions):
        raise ValueError("base_intrinsics, extrinsics and intrinsics must be all defined or all None")
    elif not all(none_field_conditions) and not all(not_none_field_conditions):
        raise ValueError(
            "cx_offset_px, cy_offset_px, img_height_px, img_width_px, pos_x_m, pos_y_m, pos_z_m, "
            "rot_x_deg, rot_z1_deg, rot_z2_deg, c1, c2, c3, c4 and pixel_aspect_ratio must be all "
            "defined or all None"
        )
    elif any(not_none_obj_conditions) and any(not_none_field_conditions):
        raise ValueError(
            "base_intrinsics, extrinsics and intrinsics must defined or cx_offset_px, cy_offset_px, img_height_px, "
            "img_width_px, pos_x_m, pos_y_m, pos_z_m, rot_x_deg, rot_z1_deg, rot_z2_deg, c1, c2, c3, c4 and "
            "pixel_aspect_ratio must be defined but not both"
        )
    if any(not_none_field_conditions):
        if validate:
            base_intrinsics = BaseIntrinsics(
                cx_offset_px=cx_offset_px,
                cy_offset_px=cy_offset_px,
                img_height_px=img_height_px,
                img_width_px=img_width_px,
            )
            extrinsics = Extrinsics(
                pos_x_m=pos_x_m,
                pos_y_m=pos_y_m,
                pos_z_m=pos_z_m,
                rot_x_deg=rot_x_deg,
                rot_z1_deg=rot_z1_deg,
                rot_z2_deg=rot_z2_deg,
            )
            intrinsics = Intrinsics(
                c1=c1,
                c2=c2,
                c3=c3,
                c4=c4,
                pixel_aspect_ratio=pixel_aspect_ratio,
            )
        else:
            base_intrinsics = BaseIntrinsics.construct(
                cx_offset_px=cx_offset_px,
                cy_offset_px=cy_offset_px,
                img_height_px=img_height_px,
                img_width_px=img_width_px,
            )
            extrinsics = Extrinsics.construct(
                pos_x_m=pos_x_m,
                pos_y_m=pos_y_m,
                pos_z_m=pos_z_m,
                rot_x_deg=rot_x_deg,
                rot_z1_deg=rot_z1_deg,
                rot_z2_deg=rot_z2_deg,
            )
            intrinsics = Intrinsics.construct(
                c1=c1,
                c2=c2,
                c3=c3,
                c4=c4,
                pixel_aspect_ratio=pixel_aspect_ratio,
            )

    if validate:
        return CamCalibration(
            type=type,
            base_intrinsics=base_intrinsics,
            extrinsics=extrinsics,
            intrinsics=intrinsics,
            id=id,
            item_ref=item_ref,
            view_ref=view_ref,
            entity_ref=entity_ref,
            source_ref=source_ref,
        )
    else:
        return CamCalibration.construct(
            type=type,
            base_intrinsics=base_intrinsics,
            extrinsics=extrinsics,
            intrinsics=intrinsics,
            id=id,
            item_ref=item_ref,
            view_ref=view_ref,
            entity_ref=entity_ref,
            source_ref=source_ref,
        )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import json
from datetime import datetime
from enum import Enum
from pathlib import Path
from types import GenericAlias
from typing import Any

from pydantic import BaseModel, create_model, model_serializer, model_validator
from typing_extensions import TYPE_CHECKING, Self

from pixano.features import BaseSchema, Item
from pixano.features.schemas.registry import _SCHEMA_REGISTRY
from pixano.features.schemas.schema_group import _SCHEMA_GROUP_TO_SCHEMA_DICT, SchemaGroup
from pixano.utils.validation import validate_and_init_create_at_and_update_at


if TYPE_CHECKING:
    from pixano.datasets.dataset import Dataset


class SchemaRelation(Enum):
    """Relation between tables.

    Attributes:
        ONE_TO_MANY: One to many relation.
        MANY_TO_ONE: Many to one relation.
        ONE_TO_ONE: One to one relation.
        MANY_TO_MANY: Many to many relation
    """

    ONE_TO_MANY = "one_to_many"
    MANY_TO_ONE = "many_to_one"
    ONE_TO_ONE = "one_to_one"
    MANY_TO_MANY = "many_to_many"


class DatasetSchema(BaseModel):
    """A dataset schema that defines the tables and the relations between them.

    Attributes:
        schemas: The mapping between the table names and their schema.
        relations: The relations between the item table and the other tables.
        groups: The groups of tables. It is filled automatically based on the schemas.
    """

    schemas: dict[str, type[BaseSchema]]
    relations: dict[str, dict[str, SchemaRelation]]
    groups: dict[SchemaGroup, set[str]] = {key: set() for key in SchemaGroup if key != SchemaGroup.SOURCE}

    def add_schema(self, table_name: str, schema: type[BaseSchema], relation_item: SchemaRelation) -> Self:
        """Add a schema to the dataset schema.

        Args:
            table_name: Name of the table to add to the dataset schema.
            schema: Schema of the table.
            relation_item: Relationship with the item schema.

        Returns:
            The dataset schema.
        """
        table_name = self.format_table_name(table_name)
        if table_name in self.schemas:
            raise ValueError(f"Table {table_name} already exists in the schemas.")
        elif not issubclass(schema, BaseSchema):
            raise ValueError(f"Schema {schema} should be a subclass of BaseSchema.")
        elif not isinstance(relation_item, SchemaRelation):
            raise ValueError(f"Invalid relation {relation_item}.")
        found_group = False
        for group, group_type in _SCHEMA_GROUP_TO_SCHEMA_DICT.items():
            if issubclass(schema, group_type):
                self.groups[group].add(table_name)
                found_group = True
                break
        if not found_group:
            raise ValueError(f"Invalid table type {schema}")
        self.schemas[table_name] = schema
        if relation_item == SchemaRelation.ONE_TO_ONE:
            self.relations[SchemaGroup.ITEM.value][table_name] = SchemaRelation.ONE_TO_ONE
            self.relations[table_name] = {SchemaGroup.ITEM.value: SchemaRelation.ONE_TO_ONE}
        elif relation_item == SchemaRelation.ONE_TO_MANY:
            self.relations[SchemaGroup.ITEM.value][table_name] = SchemaRelation.MANY_TO_ONE
            self.relations[table_name] = {SchemaGroup.ITEM.value: SchemaRelation.ONE_TO_MANY}
        elif relation_item == SchemaRelation.MANY_TO_ONE:
            self.relations[SchemaGroup.ITEM.value][table_name] = SchemaRelation.ONE_TO_MANY
            self.relations[table_name] = {SchemaGroup.ITEM.value: SchemaRelation.MANY_TO_ONE}
        elif relation_item == SchemaRelation.MANY_TO_MANY:
            self.relations[SchemaGroup.ITEM.value][table_name] = SchemaRelation.MANY_TO_MANY
            self.relations[table_name] = {SchemaGroup.ITEM.value: SchemaRelation.MANY_TO_MANY}
        return self

    @model_validator(mode="after")
    def _assign_table_groups(self) -> Self:
        for table in (
            list(self.schemas.keys())
            + list(self.relations.keys())
            + [k for v in self.relations.values() for k in v.keys()]
        ):
            # Check if table name is formatted correctly
            if table != self.format_table_name(table):
                raise ValueError(f"Table {table} should be formatted correctly.")
            # Check if table is in schemas
            elif table not in list(self.schemas.keys()):
                raise ValueError(f"Table {table} not found in schemas.")

        item_found = False
        for table, schema in self.schemas.items():
            if issubclass(schema, Item):
                if item_found:
                    raise ValueError("DatasetSchema should contain only one item schema.")
                item_found = True
            found_group = False
            for group, group_type in _SCHEMA_GROUP_TO_SCHEMA_DICT.items():
                if group == SchemaGroup.SOURCE:
                    continue
                if issubclass(schema, group_type):
                    found_group = True
                    if table not in self.groups[group]:
                        self.groups[group].add(table)
                    break
            if not found_group:
                raise ValueError(f"Invalid table type {schema}")
        if not item_found:
            raise ValueError("DatasetSchema should contain an item schema.")
        return self

    @model_validator(mode="after")
    def _check_relations(self) -> Self:
        for table, relations in self.relations.items():
            if table not in self.schemas:
                raise ValueError(f"Relation {table} not found in schemas.")
            for relation, relation_type in relations.items():
                if relation not in self.schemas:
                    raise ValueError(f"Relation {relation} not found in schemas.")
                if relation not in self.relations:
                    raise ValueError(f"Relation {relation} not found in relations.")
                if (
                    relation_type == SchemaRelation.ONE_TO_ONE
                    and not self.relations[relation][table] == SchemaRelation.ONE_TO_ONE
                ):
                    raise ValueError(f"Relation {table} -> {relation} should be one to one.")
                elif (
                    relation_type == SchemaRelation.ONE_TO_MANY
                    and not self.relations[relation][table] == SchemaRelation.MANY_TO_ONE
                ):
                    raise ValueError(f"Relation {table} -> {relation} should be one to many.")
                elif (
                    relation_type == SchemaRelation.MANY_TO_ONE
                    and not self.relations[relation][table] == SchemaRelation.ONE_TO_MANY
                ):
                    raise ValueError(f"Relation {table} -> {relation} should be many to one.")
                elif (
                    relation_type == SchemaRelation.MANY_TO_MANY
                    and not self.relations[relation][table] == SchemaRelation.MANY_TO_MANY
                ):
                    raise ValueError(f"Relation {table} -> {relation} should be many to many.")
        if SchemaGroup.ITEM.value not in self.relations:
            raise ValueError("Item schema should have relations.")
        else:
            if len(self.relations[SchemaGroup.ITEM.value]) != len(self.schemas) - 1:
                raise ValueError("Item schema should have relations with all other schemas.")
        return self

    @staticmethod
    def format_table_name(table_name: str) -> str:
        """Format table name.

        It converts the table name to lowercase and replaces spaces with underscores.

        Args:
            table_name: Table name.

        Returns:
            the formatted table name.
        """
        return table_name.lower().replace(" ", "_")

    def get_table_group(self, table_name: str) -> SchemaGroup:
        """Get the group of a table.

        Args:
            table_name: Table name.

        Returns:
            The group of the table.
        """
        for group, tables in self.groups.items():
            if table_name in tables:
                return group
        raise ValueError(f"Table {table_name} not found in groups.")

    @model_serializer
    def serialize(self) -> dict[str, dict[str, Any]]:
        """Serialize the dataset schema.

        The serialized schema is a dictionary with the following format:
        {
            "relations": {
                "item": {
                    "image": "one_to_one",
                }
            },
            "schemas": {
                "table1": {
                    "schema": "CustomItem",
                    "base_schema": "Item",
                    "fields": {
                        "id": {
                            "type": "str",
                            "collection": False
                        },
                        "split": {
                            "type": "str",
                            "collection": False
                        },
                        ...
                    }

                }
            }
        }

        Returns:
            The serialized dataset schema.
        """
        dataset_schema_json: dict[str, dict[str, Any]] = {
            "relations": {
                schema1: {schema2: relation.value for schema2, relation in relations.items()}
                for schema1, relations in self.relations.items()
            },
            "schemas": {},
            "groups": {group.value: list(schemas) for group, schemas in self.groups.items()},
        }
        for table_name, schema in self.schemas.items():
            dataset_schema_json["schemas"][table_name] = schema.serialize()
        return dataset_schema_json

    @staticmethod
    def deserialize(dataset_schema_json: dict[str, dict[str, Any]]) -> "DatasetSchema":
        """Deserialize the dataset schema.

        Args:
            dataset_schema_json: Serialized dataset schema.

        Returns:
            The dataset schema.
        """
        dataset_schema_dict: dict[str, Any] = {
            "relations": {
                schema1: {schema2: SchemaRelation(relation) for schema2, relation in relations.items()}
                for schema1, relations in dataset_schema_json["relations"].items()
            },
            "schemas": {},
            "groups": {SchemaGroup(group): set(schemas) for group, schemas in dataset_schema_json["groups"].items()},
        }
        for table_name, schema in dataset_schema_json["schemas"].items():
            dataset_schema_dict["schemas"][table_name] = BaseSchema.deserialize(schema)
        return DatasetSchema(**dataset_schema_dict)

    def to_json(self, json_fp: Path) -> None:
        """Save DatasetSchema to json file."""
        if json_fp.exists():
            old_json_content = json.loads(json_fp.read_text(encoding="utf-8"))
        else:
            old_json_content = None

        json_content = self.serialize()

        # Keep the schema field from the old json content for custom schemas.
        if old_json_content is not None:
            for table, schema in json_content["schemas"].items():
                if table not in old_json_content["schemas"]:
                    continue
                schema["schema"] = old_json_content["schemas"][table]["schema"]
        json_fp.write_text(json.dumps(json_content, indent=4), encoding="utf-8")

    @staticmethod
    def from_json(
        json_fp: Path,
    ) -> "DatasetSchema":
        """Read a dataset schema from JSON file.

        Args:
            json_fp: JSON file path

        Returns:
            The dataset schema.
        """
        schema_json = json.loads(json_fp.read_text(encoding="utf-8"))

        return DatasetSchema.deserialize(schema_json)

    @staticmethod
    def from_dataset_item(dataset_item: type["DatasetItem"]) -> "DatasetSchema":
        """Create a dataset schema from a [DatasetItem][pixano.datasets.DatasetItem].

        Args:
            dataset_item: The dataset item.

        Returns:
            The dataset schema.
        """
        item_fields = {}

        # table schemas
        dataset_schema_dict: dict[str, Any] = {}
        dataset_schema_dict["relations"] = {SchemaGroup.ITEM.value: {}}
        schemas = {}

        for field_name, field in dataset_item.model_fields.items():
            # Check if field is a generic alias (list or tuple)
            if isinstance(field.annotation, GenericAlias):
                origin = field.annotation.__origin__
                args = field.annotation.__args__

                # Check if field is list or tuple
                if origin in [list, tuple]:
                    # Categorizing list of schemas as schemas and keeping track of the relation
                    if issubclass(args[0], tuple(_SCHEMA_REGISTRY.values())):
                        schemas[field_name] = args[0]
                        dataset_schema_dict["relations"][SchemaGroup.ITEM.value][field_name] = (
                            SchemaRelation.ONE_TO_MANY
                        )
                        dataset_schema_dict["relations"][field_name] = {
                            SchemaGroup.ITEM.value: SchemaRelation.MANY_TO_ONE
                        }
                    else:
                        item_fields[field_name] = (list[args[0]], ...)  # type: ignore[valid-type]
                else:
                    # Default case: categorize as item attribute
                    item_fields[field_name] = (args[0], ...)  # type: ignore[valid-type]
            # Check if field is a schema
            elif issubclass(field.annotation, tuple(_SCHEMA_REGISTRY.values())):
                schemas[field_name] = field.annotation
                dataset_schema_dict["relations"][SchemaGroup.ITEM.value][field_name] = SchemaRelation.ONE_TO_ONE
                dataset_schema_dict["relations"][field_name] = {SchemaGroup.ITEM.value: SchemaRelation.ONE_TO_ONE}
            else:
                # Default case: item attribute
                item_fields[field_name] = (field.annotation, ...)

        CustomItem = create_model("Item", **item_fields, __base__=Item)

        schemas[SchemaGroup.ITEM.value] = CustomItem
        dataset_schema_dict["schemas"] = schemas

        return DatasetSchema(**dataset_schema_dict)


class DatasetItem(BaseModel):
    """Dataset Item.

    It is a Pydantic model that represents an item in a dataset.

    Attributes:
        id: The unique identifier of the item.
        split: The split of the item.
        created_at: The creation date of the item.
        updated_at: The last modification date of the item.
    """

    id: str
    split: str = "default"
    created_at: datetime
    updated_at: datetime

    def __init__(self, /, created_at: datetime | None = None, updated_at: datetime | None = None, **data: Any) -> None:
        """Create a new model by parsing and validating input data from keyword arguments.

        Raises [`ValidationError`][pydantic_core.ValidationError] if the input data cannot be
        validated to form a valid model.

        `self` is explicitly positional-only to allow `self` as a field name.

        Args:
            created_at: The creation date of the object.
            updated_at: The last modification date of the object.
            data: The data of the object validated by Pydantic.
        """
        created_at, updated_at = validate_and_init_create_at_and_update_at(created_at, updated_at)
        data.update({"created_at": created_at, "updated_at": updated_at})
        super().__init__(**data)

    def to_schemas_data(self, dataset_schema: DatasetSchema) -> dict[str, BaseSchema | list[BaseSchema] | None]:
        """Convert DatasetItem to schemas data.

        Args:
            dataset_schema: DatasetSchema to convert to.

        Returns:
            Schemas data.
        """
        schemas_data = {}
        item_data = {}
        for field_name in self.model_fields.keys():
            if field_name in dataset_schema.schemas:
                schemas_data[field_name] = getattr(self, field_name)
            else:
                item_data[field_name] = getattr(self, field_name)
        schemas_data[SchemaGroup.ITEM.value] = dataset_schema.schemas[SchemaGroup.ITEM.value](**item_data)
        return schemas_data

    @staticmethod
    def from_schemas_data(
        cls: "DatasetItem", schemas_data: dict[str, BaseSchema | list[BaseSchema] | None]
    ) -> "DatasetItem":
        """Create a DatasetItem from schemas data.

        Args:
            cls: The DatasetItem class.
            schemas_data: Schemas data.

        Returns:
            The created DatasetItem.
        """
        if SchemaGroup.ITEM.value not in schemas_data:
            raise ValueError("Item schema data not found.")

        schemas_data.update(schemas_data.pop(SchemaGroup.ITEM.value).model_dump())  # type: ignore[union-attr]
        return cls(**schemas_data)

    def model_dump(self, exclude_timestamps: bool = False, **kwargs: Any) -> dict[str, Any]:
        """Dump the model to a dictionary.

        Args:
            exclude_timestamps: Exclude timestamps "created_at" and "updated_at" from the model dump. Useful for
                comparing models without timestamps.
            kwargs: Arguments for pydantic `BaseModel.model_dump()`.

        Returns:
            The model dump.
        """
        model_dump = super().model_dump(**kwargs)
        if exclude_timestamps:
            model_dump.pop("created_at", None)
            model_dump.pop("updated_at", None)
            for k, value in model_dump.items():
                if isinstance(value, dict):
                    value.pop("created_at", None)
                    value.pop("updated_at", None)
                elif isinstance(value, list):  # Only one level deep.
                    for item in value:
                        if isinstance(item, dict):
                            item.pop("created_at", None)
                            item.pop("updated_at", None)
        return model_dump

    def model_copy(self, *, dataset: "Dataset", deep=False) -> Self:
        """Returns a copy of the model.

        Args:
            dataset: The dataset where the DatasetItem belongs.
            deep: Set to `True` to make a deep copy of the model.

        Returns:
            New model instance.
        """
        # Actual copy done by each schema to call our own model_copy method
        data: dict[str, BaseSchema | list[BaseSchema] | None] = self.to_schemas_data(dataset.schema)
        copied_data: dict[str, BaseSchema | list[BaseSchema] | None] = {}
        for key, value in data.items():
            if isinstance(value, list):
                copied_data[key] = [item.model_copy(deep=deep) for item in value]
            elif value is not None:
                copied_data[key] = value.model_copy(deep=deep)
            else:
                copied_data[key] = None
        copy_item = self.from_schemas_data(self.__class__, copied_data)  # type: ignore[arg-type]
        return copy_item

    @classmethod
    def to_dataset_schema(cls) -> DatasetSchema:
        """Convert a DatasetItem to a DatasetSchema."""
        return DatasetSchema.from_dataset_item(cls)

    @staticmethod
    def from_dataset_schema(dataset_schema: DatasetSchema, exclude_embeddings: bool = True) -> type["DatasetItem"]:
        """Create a dataset item model based on the schema.

        Args:
            dataset_schema: The dataset schema.
            exclude_embeddings: Exclude embeddings from the dataset item model to reduce the size.

        Returns:
            The dataset item model
        """
        item_type = dataset_schema.schemas[SchemaGroup.ITEM.value]
        fields: dict[str, Any] = {}

        if dataset_schema.relations != {} and SchemaGroup.ITEM.value in dataset_schema.relations:
            for schema, relation in dataset_schema.relations[SchemaGroup.ITEM.value].items():
                if exclude_embeddings and schema in dataset_schema.groups[SchemaGroup.EMBEDDING]:
                    continue
                # Add default value in case an item does not have a specific view or entity.
                schema_type = dataset_schema.schemas[schema]
                if relation == SchemaRelation.ONE_TO_MANY:
                    fields[schema] = (list[schema_type], [])  # type: ignore[valid-type]
                else:
                    fields[schema] = (schema_type | None, None)

        for field_name, field in item_type.model_fields.items():
            # No default value as all items metadata should be retrieved.
            fields[field_name] = (field.annotation, ...)

        CustomDatasetItem = create_model(
            "DatasetItem",
            **fields,
            __base__=DatasetItem,
        )
        return CustomDatasetItem

    @classmethod
    def get_sub_dataset_item(cls, selected_fields: list[str]) -> type[Self]:
        """Create a new dataset item based on the selected fields of the original dataset
        item.

        Note:
            The id and split fields are always included in the sub dataset item.

        Note:
            The sub dataset item does not have the methods and config of the original
            dataset item.

        Args:
            selected_fields: The selected fields.

        Returns:
            The sub dataset item.
        """
        fields = {}
        for field_name, field in cls.model_fields.items():
            if field_name in selected_fields or field_name in ["id", "split"]:
                if isinstance(field.annotation, GenericAlias):
                    origin = field.annotation.__origin__
                    args = field.annotation.__args__

                    # Check if field is list or tuple
                    if origin is tuple:
                        fields[field_name] = (origin[args[0], ...], field.default)  # type: ignore[index]
                    else:
                        fields[field_name] = (field.annotation, field.default)
                else:
                    fields[field_name] = (field.annotation, field.default)

        SubDatasetItem: type[DatasetItem] = create_model(
            cls.__name__,
            **fields,
            __base__=DatasetItem,
        )

        return SubDatasetItem



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Any, Literal

import numpy as np
from pydantic import model_validator
from typing_extensions import Self

from pixano.features.utils import boxes as bbox_utils
from pixano.utils import issubclass_strict

from ...types.schema_reference import EntityRef, ItemRef, SourceRef, ViewRef
from ..registry import _register_schema_internal
from .annotation import Annotation
from .compressed_rle import CompressedRLE


@_register_schema_internal
class BBox(Annotation):
    """Bounding box using coordinates in xyxy or xywh format.

    Attributes:
        coords: List of coordinates in given format.
        format: Coordinates format, 'xyxy' or 'xywh'.
        is_normalized: True if coordinates are normalized to image size.
        confidence: Bounding box confidence if predicted. -1 if not predicted.
    """

    coords: list[float]
    format: str
    is_normalized: bool
    confidence: float = -1.0

    @model_validator(mode="after")
    def _validate_fields(self) -> Self:
        if len(self.coords) != 4:
            raise ValueError("Bounding box coordinates must have 4 elements.")
        elif not all(coord >= 0 for coord in self.coords):
            raise ValueError("Bounding box coordinates must be positive.")
        elif self.is_normalized and not all(0 <= coord <= 1 for coord in self.coords):
            raise ValueError("Normalized bounding box coordinates must be in [0, 1] range.")
        elif (self.confidence < 0 or self.confidence > 1) and not self.confidence == -1:
            raise ValueError("Bounding box confidence must be in [0, 1] range or -1.")
        elif self.format not in ["xyxy", "xywh"]:
            raise ValueError("Bounding box format must be 'xyxy' or 'xywh'.")
        return self

    @classmethod
    def none(cls) -> Self:
        """Utility function to get a `None` equivalent.
        Should be removed as soon as Lance manages `None` value.

        Returns:
            "None" `BBox`.
        """
        return cls(
            id="",
            item=ItemRef.none(),
            view=ViewRef.none(),
            entity=EntityRef.none(),
            coords=[0.0, 0.0, 0.0, 0.0],
            format="xywh",
            is_normalized=True,
            confidence=-1,
        )

    @property
    def xyxy_coords(self) -> list[float]:
        """Return the bounding box xyxy coordinates.

        Returns:
            Coordinates in xyxy format.
        """
        return self.coords if self.format == "xyxy" else bbox_utils.xywh_to_xyxy(self.coords)

    @property
    def xywh_coords(self) -> list[float]:
        """Return the bounding box xywh coordinates.

        Returns:
            Coordinates in xywh format.
        """
        return self.coords if self.format == "xywh" else bbox_utils.xyxy_to_xywh(self.coords)

    def to_xyxy(self) -> Self:
        """Return the bounding box in xyxy format.

        Returns:
            Bounding box in xyxy format.
        """
        return BBox(
            coords=self.xyxy_coords,
            format="xyxy",
            is_normalized=self.is_normalized,
            confidence=self.confidence,
        )

    def to_xywh(self) -> Self:
        """Return the bounding box in xywh format.

        Returns:
            Bounding box in xyxy format.
        """
        return BBox(
            coords=self.xywh_coords,
            format="xywh",
            is_normalized=self.is_normalized,
            confidence=self.confidence,
        )

    def normalize(self, height: int, width: int) -> Self:
        """Return the bounding box with coordinates normalized relatively to the image size.

        Args:
            height: Image height.
            width: Image width.

        Returns:
            Bounding box with coordinates normalized relatively to the image size.
        """
        return BBox(
            coords=bbox_utils.normalize_coords(self.coords, height, width),
            format=self.format,
            is_normalized=True,
            confidence=self.confidence,
        )

    def denormalize(self, height: int, width: int) -> Self:
        """Return the bounding box with coordinates denormalized relatively to the image size.

        Args:
            height: Image height.
            width: Image width.

        Returns:
            Bounding box with coordinates denormalized relatively to the image size.
        """
        return BBox(
            coords=bbox_utils.denormalize_coords(self.coords, height, width),
            format=self.format,
            is_normalized=False,
            confidence=self.confidence,
        )

    @staticmethod
    def from_xyxy(
        xyxy: list[float],
        **kwargs: Any,
    ) -> "BBox":
        """Create a bounding box using normalized xyxy coordinates.

        Args:
            xyxy: List of coordinates in xyxy format.
            kwargs: Additional arguments.

        Returns:
            The bounding box.
        """
        return BBox(
            coords=xyxy,
            format="xyxy",
            **kwargs,
        )

    @staticmethod
    def from_xywh(
        xywh: list[float],
        **kwargs: Any,
    ) -> "BBox":
        """Create a bounding box using normalized xywh coordinates.

        Args:
            xywh: List of coordinates in xywh format.
            kwargs: Additional arguments.

        Returns:
            The bounding box.
        """
        return BBox(
            coords=xywh,
            format="xywh",
            **kwargs,
        )

    @staticmethod
    def from_mask(mask: np.ndarray, **kwargs: Any) -> "BBox":
        """Create a bounding box using a NumPy array mask.

        Args:
            mask: NumPy array mask.
            kwargs: Additional arguments.

        Returns:
            The bounding box.
        """
        return BBox.from_xywh(
            xywh=bbox_utils.mask_to_bbox(mask),
            is_normalized=True,
            **kwargs,
        )

    @staticmethod
    def from_rle(
        rle: CompressedRLE,
        **kwargs: Any,
    ) -> "BBox":
        """Create a bounding box using a RLE mask.

        Args:
            rle: RLE mask.
            kwargs: Additional arguments.

        Returns:
            The bounding box.
        """
        return BBox.from_mask(mask=rle.to_mask(), **kwargs)


@_register_schema_internal
class BBox3D(Annotation):
    """A 3D bounding Box.

    Attributes:
        coords: List of coordinates in given format.
        format: Coordinates format, 'xyzxyz' or 'xyzwhd'.
        heading: Orientation of the bounding box.
        is_normalized: True if coordinates are normalized to image size.
        confidence: Bounding box confidence if predicted. -1 if not predicted.
    """

    coords: list[float]
    format: str
    heading: list[float]
    is_normalized: bool
    confidence: float = -1.0

    @model_validator(mode="after")
    def _validate_fields(self):
        if len(self.coords) != 6:
            raise ValueError("3D Bounding box coordinates must have 6 elements.")
        elif not all(coord >= 0 for coord in self.coords):
            raise ValueError("Bounding box coordinates must be positive.")
        elif self.is_normalized and not all(0 <= coord <= 1 for coord in self.coords):
            raise ValueError("Normalized bounding box coordinates must be in [0, 1] range.")
        elif (self.confidence < 0 or self.confidence > 1) and not self.confidence == -1:
            raise ValueError("Bounding box confidence must be in [0, 1] range or -1.")
        elif self.format not in ["xyzxyz", "xyzwhd"]:
            raise ValueError("Bounding box format must be 'xyzxyz' or 'xyzwhd'.")
        return self

    @classmethod
    def none(cls) -> Self:
        """Utility function to get a `None` equivalent.
        Should be removed as soon as Lance manages `None` value.

        Returns:
            "None" BBox3D.
        """
        return cls(
            id="",
            item=ItemRef.none(),
            view=ViewRef.none(),
            entity=EntityRef.none(),
            coords=[0, 0, 0, 0, 0, 0],
            format="xyzwhd",
            heading=[0, 0, 0],
            is_normalized=True,
            confidence=-1,
        )


def is_bbox(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `BBox` or a subclass of `BBox`."""
    return issubclass_strict(cls, BBox, strict)


def is_bbox3d(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `BBox3D` or subclass of `BBox3D`."""
    return issubclass_strict(cls, BBox3D, strict)


def create_bbox(
    coords: list[float],
    format: Literal["xyxy", "xywh"],
    is_normalized: bool,
    confidence: float = -1,
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    view_ref: ViewRef = ViewRef.none(),
    entity_ref: EntityRef = EntityRef.none(),
    source_ref: SourceRef = SourceRef.none(),
) -> BBox:
    """Create a `BBox` instance.

    Args:
        coords: List of coordinates in given format.
        format: Coordinates format, 'xyxy' or 'xywh'.
        is_normalized: True if coordinates are normalized to image size.
        confidence: Bounding box confidence if predicted.
        id: BBox ID.
        item_ref: Item reference.
        view_ref: View reference.
        entity_ref: Entity reference.
        source_ref: Source reference.

    Returns:
        The created `BBox` instance.
    """
    return BBox(
        id=id,
        item_ref=item_ref,
        view_ref=view_ref,
        entity_ref=entity_ref,
        source_ref=source_ref,
        coords=coords,
        format=str(format),
        is_normalized=is_normalized,
        confidence=confidence,
    )


def create_bbox3d(
    coords: list[float],
    format: Literal["xyzxyz", "xyzwhd"],
    heading: list[float],
    is_normalized: bool,
    confidence: float = -1.0,
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    view_ref: ViewRef = ViewRef.none(),
    entity_ref: EntityRef = EntityRef.none(),
    source_ref: SourceRef = SourceRef.none(),
) -> BBox3D:
    """Create a `BBox3D` instance.

    Args:
        coords: The 3D position coordinates.
        format: Coordinates format, 'xyzxyz' or 'xyzwhd'.
        heading: The orientation.
        is_normalized: True if coordinates are normalized to image size.
        confidence: Bounding box confidence if predicted.
        id: BBox3D ID.
        item_ref: Item reference.
        view_ref: View reference.
        entity_ref: Entity reference.
        source_ref: Source reference.

    Returns:
        The created `BBox3D` instance.
    """
    return BBox3D(
        id=id,
        item_ref=item_ref,
        view_ref=view_ref,
        entity_ref=entity_ref,
        source_ref=source_ref,
        coords=coords,
        format=str(format),
        heading=heading,
        is_normalized=is_normalized,
        confidence=confidence,
    )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pydantic import ConfigDict, field_validator
from typing_extensions import TypeVar

from pixano.datasets import Dataset
from pixano.features import Embedding, SchemaGroup, is_embedding

from .base_schema import BaseSchemaModel
from .table_info import TableInfo


T = TypeVar("T", bound=Embedding)


class EmbeddingModel(BaseSchemaModel[Embedding]):
    """Model for the [Embedding][pixano.features.Embedding] schema."""

    model_config = ConfigDict(
        validate_assignment=True,
        json_schema_extra={
            "examples": [
                {
                    "id": "embedding_orange_cats",
                    "table_info": {"group": "embeddings", "name": "cat_embeddings", "base_schema": "ViewEmbedding"},
                    "data": {
                        "item_ref": {"name": "item", "id": "1"},
                        "view_ref": {"name": "image", "id": "orange_cats"},
                        "vector": [0.0, 1.0, 2.0, -1.0, -2.0, 0.0],
                        "shape": [6],
                    },
                }
            ]
        },
    )

    @field_validator("table_info")
    @classmethod
    def _validate_table_info(cls, value: TableInfo) -> TableInfo:
        """Validate table info."""
        if value.group != SchemaGroup.EMBEDDING.value:
            raise ValueError(f"Table info group must be {SchemaGroup.EMBEDDING.value}.")
        return value

    def to_row(self, dataset: Dataset) -> Embedding:
        """Create an [Embedding][pixano.features.Embedding] from the model."""
        if not is_embedding(dataset.schema.schemas[self.table_info.name]):
            raise ValueError(f"Schema type must be a subclass of {Embedding.__name__}.")
        return super().to_row(dataset)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .errors import DatasetAccessError, DatasetPaginationError, DatasetWriteError
from .integrity import (
    check_dataset_integrity,
    check_table_integrity,
    get_integry_checks_from_schemas,
    handle_integrity_errors,
)
from .labels import category_id, category_name, coco_ids_80to91
from .mosaic import mosaic
from .video import create_video_preview


__all__ = [
    "DatasetAccessError",
    "DatasetPaginationError",
    "DatasetWriteError",
    "check_dataset_integrity",
    "check_table_integrity",
    "create_video_preview",
    "coco_ids_80to91",
    "category_id",
    "category_name",
    "get_integry_checks_from_schemas",
    "handle_integrity_errors",
    "mosaic",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .base_schema import BaseSchema


_PIXANO_SCHEMA_REGISTRY: dict[str, type[BaseSchema]] = {"BaseSchema": BaseSchema}
_SCHEMA_REGISTRY: dict[str, type[BaseSchema]] = {"BaseSchema": BaseSchema}


def _add_schema_to_registry(schema: type[BaseSchema], registry: dict[str, type[BaseSchema]]) -> None:
    if not issubclass(schema, BaseSchema):
        raise ValueError(f"Schema {schema} must be a subclass of BaseSchema")

    schema_name = schema.__name__
    if schema_name in registry:
        raise ValueError(f"Schema {schema} already registered")
    registry[schema_name] = schema
    return None


def register_schema(cls: type[BaseSchema]):
    """Class decorator to register a schema.

    Args:
        cls: The schema to register.
    """
    _add_schema_to_registry(cls, _SCHEMA_REGISTRY)
    return cls


def _register_schema_internal(cls):
    _add_schema_to_registry(cls, _SCHEMA_REGISTRY)
    _add_schema_to_registry(cls, _PIXANO_SCHEMA_REGISTRY)
    return cls



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from __future__ import annotations

import json
from pathlib import Path
from typing import Literal, overload

from pydantic import BaseModel, field_serializer, field_validator

from pixano.datasets.workspaces import WorkspaceType
from pixano.features import Image


class DatasetInfo(BaseModel):
    """Information of a dataset.

    Attributes:
        id: Dataset ID. Must be unique.
        name: Dataset name.
        description: Dataset description.
        estimated_size: Dataset estimated size.
        preview: Path to a preview thumbnail.
        workspace: Workspace type.
    """

    id: str = ""
    name: str = ""
    description: str = ""
    size: str = "Unknown"
    preview: str = ""
    workspace: WorkspaceType = WorkspaceType.UNDEFINED

    @field_serializer("workspace")
    def serialize_workspace(self, workspace: WorkspaceType):
        """Dump workspace as string value, not enum."""
        return workspace.value

    @field_validator("id", mode="after")
    @classmethod
    def _id_validator(cls, v: str) -> str:
        if " " in v:
            raise ValueError("id must not contain spaces")
        return v

    def to_json(self, json_fp: Path) -> None:
        """Writes the DatasetInfo object to a JSON file.

        Args:
            json_fp: The path to the file where the DatasetInfo object
                will be written.
        """
        model_dumped = self.model_dump()
        json_fp.write_text(json.dumps(model_dumped, indent=4), encoding="utf-8")

    @staticmethod
    def from_json(
        json_fp: Path,
    ) -> "DatasetInfo":
        """Read DatasetInfo from JSON file.

        Args:
            json_fp: JSON file path.

        Returns:
            the dataset info object.
        """
        info_json = json.loads(json_fp.read_text(encoding="utf-8"))

        info_json["workspace"] = (
            WorkspaceType(info_json["workspace"]) if "workspace" in info_json else WorkspaceType.UNDEFINED
        )
        info = DatasetInfo.model_validate(info_json)

        return info

    @overload
    @staticmethod
    def load_directory(
        directory: Path,
        return_path: Literal[False] = False,
    ) -> list["DatasetInfo"]: ...
    @overload
    @staticmethod
    def load_directory(
        directory: Path,
        return_path: Literal[True],
    ) -> list[tuple["DatasetInfo", Path]]: ...
    @staticmethod
    def load_directory(
        directory: Path,
        return_path: bool = False,
    ) -> list[tuple["DatasetInfo", Path]] | list["DatasetInfo"]:
        """Load list of DatasetInfo from directory.

        Args:
            directory: Directory to load.
            return_path: Return the paths of the datasets.

        Returns:
            The list of DatasetInfo and the paths of the datasets.
        """
        library: list[DatasetInfo] | list[tuple[DatasetInfo, Path]] = []

        # Browse directory
        for json_fp in sorted(directory.glob("*/info.json")):
            info: DatasetInfo = DatasetInfo.from_json(json_fp)
            try:
                info.preview = Image.open_url(
                    str(json_fp.parent / "previews/dataset_preview.jpg"),
                    Path("/"),
                )  # TODO choose correct preview name / path / extension
            except Exception:  # TODO: specify exception URL and Value
                info.preview = ""
            if return_path:
                library.append((info, json_fp.parent))  #  type: ignore[arg-type]
            else:
                library.append(info)

        if library == []:
            raise FileNotFoundError(f"No dataset found in {directory}.")

        return library

    @overload
    @staticmethod
    def load_id(id: str, directory: Path, return_path: Literal[False] = False) -> "DatasetInfo": ...
    @overload
    @staticmethod
    def load_id(id: str, directory: Path, return_path: Literal[True] = True) -> tuple["DatasetInfo", Path]: ...
    @staticmethod
    def load_id(id: str, directory: Path, return_path: bool = False) -> tuple["DatasetInfo", Path] | "DatasetInfo":
        """Load a specific DatasetInfo from directory.

        Args:
            id: The ID of the dataset to load.
            directory: Directory to load.
            return_path: Return the path of the dataset.

        Returns:
            The DatasetInfo.
        """
        for json_fp in directory.glob("*/info.json"):
            info = DatasetInfo.from_json(json_fp)
            if info.id == id:
                try:
                    info.preview = Image.open_url(
                        str(json_fp.parent / "previews/dataset_preview.jpg"),
                        json_fp.parent / "media",
                    )  # TODO choose correct preview name / path / extension
                except ValueError:
                    info.preview = ""
                return (info, json_fp.parent) if return_path else info
        raise FileNotFoundError(f"No dataset found with ID {id}")



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pixano.utils import issubclass_strict

from ...types.schema_reference import ItemRef, ViewRef
from ..registry import _register_schema_internal
from .view import View


@_register_schema_internal
class Text(View):
    """Text view.

    Attributes:
        content: The text content.
    """

    content: str


def is_text(cls: type, strict: bool = False) -> bool:
    """Check if the given class is `Text` or a subclass of `Text`."""
    return issubclass_strict(cls, Text, strict)


def create_text(
    content: str,
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    parent_ref: ViewRef = ViewRef.none(),
) -> Text:
    """Create a `Text` instance.

    Args:
        content: The text content.
        id: Text ID.
        item_ref: Item reference.
        parent_ref: Parent view reference.

    Returns:
        The created `Text` instance.
    """
    return Text(id=id, item_ref=item_ref, parent_ref=parent_ref, content=content)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import os
import re
from collections import OrderedDict
from pathlib import Path
from typing import Any, Sequence


def natural_key(string: str) -> list:
    """Return key for string natural sort.

    Args:
        string: Input string.

    Returns:
        Sort key.
    """
    return [int(s) if s.isdecimal() else s for s in re.split(r"(\d+)", string)]


def estimate_folder_size(folder_path: Path) -> str:
    """Estimate a folder size and return it as a human-readable string.

    Args:
        folder_path: Folder path.

    Returns:
        Folder size as a human-readable string.
    """
    # Estimate size
    total_size = 0.0
    for dirpath, _, filenames in os.walk(folder_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    # Format size
    i = 0
    suffixes = ["B", "KB", "MB", "GB", "TB", "PB"]
    while total_size >= 1024 and i < len(suffixes) - 1:
        total_size /= 1024.0
        i += 1
    f = (f"{total_size:.2f}").rstrip("0").rstrip(".")
    readable_size = f"{f} {suffixes[i]}"

    return readable_size


def get_super_type_from_dict(sub_type: type, dict_types: dict[str, type]) -> type | None:
    """Get the first super type in a dictionary of types for the given type.

    Args:
        sub_type: Sub type to find the super type for.
        dict_types: Dictionary of types.

    Returns:
        Super type if found, None otherwise.
    """
    if sub_type in dict_types.values():
        return sub_type

    sup_type = None
    for dict_type in dict_types.values():
        if issubclass(sub_type, dict_type):
            sup_type = dict_type
            break

    if sup_type is None:
        return None

    found_type = True
    while found_type:
        found_type = False
        for dict_type in dict_types.values():
            if issubclass(sub_type, dict_type) and issubclass(dict_type, sup_type) and dict_type is not sup_type:
                sup_type = dict_type
                found_type = True
                break

    return sup_type


def to_sql_list(ids: str | Sequence[str] | set[str]) -> str:
    """Convert a list of IDs to a SQL-friendly string.

    Args:
        ids: List of IDs.

    Returns:
        SQL-friendly string of IDs.
    """
    if isinstance(ids, str):
        return f"('{ids}')"
    elif len(ids) == 0:
        raise ValueError("IDs must not be empty.")
    else:
        for id in ids:
            if not isinstance(id, str):
                raise ValueError("IDs must be strings.")
    ids = list(dict.fromkeys(ids))  # Keep order and remove duplicates
    if len(ids) == 1:
        return f"('{ids.pop()}')"
    return str(tuple(ids))


def fn_sort_dict(dict_: dict[str, Any], order_by: list[str], descending: list[bool]) -> tuple[Any, ...]:
    """Function to sort a dictionary by multiple keys in different orders.

    Args:
        dict_: Dictionary to sort.
        order_by: List of keys to sort by.
        descending: List of booleans indicating the order for each key.
    """
    key: list[Any] = []
    for col, desc in zip(order_by, descending, strict=True):
        value = dict_.get(col)
        if desc:
            if isinstance(value, (int, float)):
                key.append(-value)
            elif isinstance(value, str):
                key.append("".join(chr(255 - ord(c)) for c in value))
            elif value is bool:
                key.append(not value)
            elif value is None:
                key.append(None)
            else:
                raise ValueError(
                    f"Cannot sort by {type(value)} in descending order. "
                    "Please use open an issue if you need this feature."
                )
        else:
            key.append(value)
    return tuple(key)


def unique_list(sequence: Sequence[Any]) -> list[Any]:
    """Select unique elements in a list while keeping order.

    Args:
        sequence: Input sequence.

    Returns:
        List of unique elements.
    """
    return list(OrderedDict.fromkeys(sequence))



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Annotated

from fastapi import APIRouter, Depends, Query

from pixano.app.models.embeddings import EmbeddingModel
from pixano.app.settings import Settings, get_settings
from pixano.features.schemas.schema_group import SchemaGroup

from .utils import (
    create_row_handler,
    create_rows_handler,
    delete_row_handler,
    delete_rows_handler,
    get_row_handler,
    get_rows_handler,
    update_row_handler,
    update_rows_handler,
)


router = APIRouter(prefix="/embeddings", tags=["Embeddings"])


@router.get("/{dataset_id}/{table}/", response_model=list[EmbeddingModel])
async def get_embeddings(
    dataset_id: str,
    table: str,
    settings: Annotated[Settings, Depends(get_settings)],
    ids: list[str] | None = Query(None),
    limit: int | None = None,
    skip: int = 0,
    where: str | None = None,
    item_ids: list[str] | None = Query(None),
) -> list[EmbeddingModel]:
    """Get embeddings from a table of a dataset.

    They can be filtered by IDs, item IDs, a where clause or paginated.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        settings: App settings.
        ids: IDs of the embeddings.
        limit: Limit number of embeddings.
        skip: Skip number of embeddings.
        where: Where clause.
        item_ids: Item IDs.

    Returns:
        List of embeddings.
    """
    return await get_rows_handler(
        dataset_id=dataset_id,
        group=SchemaGroup.EMBEDDING,
        table=table,
        settings=settings,
        where=where,
        ids=ids,
        item_ids=item_ids,
        limit=limit,
        skip=skip,
    )


@router.get("/{dataset_id}/{table}/{id}", response_model=EmbeddingModel)
async def get_embedding(
    dataset_id: str, table: str, id: str, settings: Annotated[Settings, Depends(get_settings)]
) -> EmbeddingModel:
    """Get an embedding from a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        id: ID of the embedding.
        settings: App settings.

    Returns:
        The embedding.
    """
    return await get_row_handler(dataset_id, SchemaGroup.EMBEDDING, table, id, settings)


@router.post("/{dataset_id}/{table}/", response_model=list[EmbeddingModel])
async def create_embeddings(
    dataset_id: str,
    table: str,
    embeddings: list[EmbeddingModel],
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[EmbeddingModel]:
    """Add embeddings in a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        embeddings: Embeddings to add.
        settings: App settings.

    Returns:
        List of embeddings added.
    """
    return await create_rows_handler(dataset_id, SchemaGroup.EMBEDDING, table, embeddings, settings)


@router.post("/{dataset_id}/{table}/{id}", response_model=EmbeddingModel)
async def create_embedding(
    dataset_id: str,
    table: str,
    id: str,
    embedding: EmbeddingModel,
    settings: Annotated[Settings, Depends(get_settings)],
) -> EmbeddingModel:
    """Update an embedding in a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        id: ID of the embedding.
        embedding: Embedding to update.
        settings: App settings.

    Returns:
        The embedding updated.
    """
    return await create_row_handler(dataset_id, SchemaGroup.EMBEDDING, table, id, embedding, settings)


@router.put("/{dataset_id}/{table}/{id}", response_model=EmbeddingModel)
async def update_embedding(
    dataset_id: str,
    table: str,
    id: str,
    embedding: EmbeddingModel,
    settings: Annotated[Settings, Depends(get_settings)],
) -> EmbeddingModel:
    """Update an embedding in a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        id: ID of the embedding.
        embedding: Embedding to update.
        settings: App settings.

    Returns:
        The embedding updated.
    """
    return await update_row_handler(dataset_id, SchemaGroup.EMBEDDING, table, id, embedding, settings)


@router.put("/{dataset_id}/{table}/", response_model=list[EmbeddingModel])
async def update_embeddings(
    dataset_id: str,
    table: str,
    embeddings: list[EmbeddingModel],
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[EmbeddingModel]:
    """Update embeddings in a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        embeddings: Embeddings to update.
        settings: App settings.

    Returns:
        List of embeddings updated.
    """
    return await update_rows_handler(dataset_id, SchemaGroup.EMBEDDING, table, embeddings, settings)


@router.delete("/{dataset_id}/{table}/{id}")
async def delete_embedding(
    dataset_id: str, table: str, id: str, settings: Annotated[Settings, Depends(get_settings)]
) -> None:
    """Delete an embedding from a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        id: ID of the embedding to delete.
        settings: App settings.
    """
    return await delete_row_handler(dataset_id, SchemaGroup.EMBEDDING, table, id, settings)


@router.delete("/{dataset_id}/{table}/")
async def delete_embeddings(
    dataset_id: str,
    table: str,
    ids: Annotated[list[str], Query()],
    settings: Annotated[Settings, Depends(get_settings)],
) -> None:
    """Delete embeddings from a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        ids: IDs of the embeddings to delete.
        settings: App settings.
    """
    return await delete_rows_handler(dataset_id, SchemaGroup.EMBEDDING, table, ids, settings)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from lancedb.pydantic import LanceModel
from pydantic import ConfigDict

from pixano.utils import issubclass_strict


class BaseType(LanceModel):
    """Base class for all Pixano types.

    This class should be inherited by all Pixano types. [BaseSchema][pixano.features.BaseSchema]s can only have a
    primitive type or a `BaseType` attribute.
    """

    model_config = ConfigDict(validate_assignment=True)
    pass


def is_base_type(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `BaseType` or subclass of `BaseType`."""
    return issubclass_strict(cls, BaseType, strict)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pixano.features.schemas.schema_group import SchemaGroup

from .annotations import AnnotationModel
from .base_schema import BaseSchemaModel
from .embeddings import EmbeddingModel
from .entities import EntityModel
from .items import ItemModel
from .sources import SourceModel
from .views import ViewModel


_SCHEMA_GROUP_TO_SCHEMA_MODEL_DICT: dict[SchemaGroup, type[BaseSchemaModel]] = {
    SchemaGroup.EMBEDDING: EmbeddingModel,
    SchemaGroup.ITEM: ItemModel,
    SchemaGroup.ENTITY: EntityModel,
    SchemaGroup.ANNOTATION: AnnotationModel,
    SchemaGroup.VIEW: ViewModel,
    SchemaGroup.SOURCE: SourceModel,
}



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import warnings

from pydantic import model_validator
from typing_extensions import Self

from pixano.utils import issubclass_strict

from ...types.schema_reference import AnnotationRef, EntityRef, ItemRef, SourceRef, ViewRef
from ..registry import _register_schema_internal
from .annotation import Annotation


@_register_schema_internal
class TextSpan(Annotation):
    """Designation of a Text-Span in a text, especially in the
       use-case of Named-Entity Recognition on a textual annotation
       having a str 'content' attribute .

    Attributes:
        mention: text-span assembled mention.
        spans_start: List of start offsets of the spans in the text.
        spans_end: List of end offsets of the spans in the text.
        annotation_ref: Annotation reference toward an textual Annotation having a
          str 'content' attribute
    """

    mention: str
    spans_start: list[int]
    spans_end: list[int]
    annotation_ref: AnnotationRef

    @model_validator(mode="after")
    def _validate_fields(self) -> Self:
        if len(self.spans_start) != len(self.spans_end):
            raise ValueError("Spans offset lists should have the same length")
        if len(self.spans_start) == 0:
            warnings.warn("To ground a TextSpan in a text, spans offsets should not be empty", category=UserWarning)
        if not all(span[0] >= 0 and span[1] >= 0 for span in zip(self.spans_start, self.spans_end)):
            raise ValueError("Spans offset must be positive")
        if not all(span[0] <= span[1] for span in zip(self.spans_start, self.spans_end)):
            raise ValueError("End offsets must be greater or equal to their corresponding start offset")
        return self

    @classmethod
    def none(cls) -> Self:
        """Utility function to get a `None` equivalent.
        Should be removed as soon as Lance manages `None` value.

        Returns:
            "None" TextSpan.
        """
        return cls(
            id="",
            item_ref=ItemRef.none(),
            view_ref=ViewRef.none(),
            entity_ref=EntityRef.none(),
            mention="",
            spans_start=[],
            spans_end=[],
            annotation_ref=AnnotationRef.none(),
        )

    @property
    def spans(self) -> list[tuple[int, int]]:
        """Get the list of zipped spans offsets (starts and ends)."""
        return list(zip(self.spans_start, self.spans_end))

    @property
    def spans_length(self) -> list[int]:
        """Get the computed list of spans lengths."""
        return [e - s for s, e in zip(self.spans_start, self.spans_end)]


@_register_schema_internal
class Relation(Annotation):
    """Observation of a relation between two annotations,
       for instance between text-spans in a text.

    Attributes:
        predicate: type of relation, as in semantic-web (OWL, RDF, etc)
        subject_ref: annotation_ref to the subject Annotation (eg TextSpan)
        object_ref: annotation_ref to the object Annotation (eg TextSpan)
    """

    predicate: str
    subject_ref: AnnotationRef
    object_ref: AnnotationRef

    @classmethod
    def none(cls) -> Self:
        """Utility function to get a `None` equivalent.
        Should be removed as soon as Lance manages `None` value.

        Returns:
            "None" Relation.
        """
        return cls(
            id="",
            predicate="",
            item_ref=ItemRef.none(),
            view_ref=ViewRef.none(),
            entity_ref=EntityRef.none(),
            subject_ref=AnnotationRef.none(),
            object_ref=AnnotationRef.none(),
        )


def is_text_span(cls: type, strict: bool = False) -> bool:
    """Check if a class is a TextSpan or subclass of TextSpan."""
    return issubclass_strict(cls, TextSpan, strict)


def is_relation(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `Relation` or subclass of `Relation`."""
    return issubclass_strict(cls, Relation, strict)


def create_text_span(
    mention: str,
    spans_start: list[int],
    spans_end: list[int],
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    view_ref: ViewRef = ViewRef.none(),
    entity_ref: EntityRef = EntityRef.none(),
    source_ref: SourceRef = SourceRef.none(),
    annotation_ref: AnnotationRef = AnnotationRef.none(),
) -> TextSpan:
    """Create a TextSpan instance.

    Args:
        mention: text-span observed mention.
        spans_start: List of start offsets of the spans in the text.
        spans_end: List of end offsets of the spans in the text.
        id: TextSpan ID.
        item_ref: Item reference.
        view_ref: View reference.
        entity_ref: Entity reference.
        source_ref: Source reference.
        annotation_ref: Annotation reference toward an textual Annotation having a
          str 'content' attribute

    Returns:
        The created `TextSpan` instance.
    """
    return TextSpan(
        mention=mention,
        spans_start=spans_start,
        spans_end=spans_end,
        annotation_ref=annotation_ref,
        id=id,
        item_ref=item_ref,
        view_ref=view_ref,
        entity_ref=entity_ref,
        source_ref=source_ref,
    )


def create_relation(
    predicate: str,
    subject_ref: AnnotationRef = AnnotationRef.none(),
    object_ref: AnnotationRef = AnnotationRef.none(),
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    view_ref: ViewRef = ViewRef.none(),
    entity_ref: EntityRef = EntityRef.none(),
    source_ref: SourceRef = SourceRef.none(),
) -> Relation:
    """Create a `Relation` instance.

    Args:
        predicate: type of relation
        subject_ref: annotation_ref to the subject TextSpan
        object_ref: annotation_ref to the object TextSpan
        id: Relation ID.
        item_ref: Item reference.
        view_ref: View reference.
        entity_ref: Entity reference.
        source_ref: Source reference.

    Returns:
        The created `Relation` instance.
    """
    return Relation(
        predicate=predicate,
        subject_ref=subject_ref,
        object_ref=object_ref,
        id=id,
        item_ref=item_ref,
        view_ref=view_ref,
        entity_ref=entity_ref,
        source_ref=source_ref,
    )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pydantic import model_validator
from typing_extensions import Self

from pixano.utils import issubclass_strict

from ...types.schema_reference import EntityRef, ItemRef, SourceRef, ViewRef
from ..registry import _register_schema_internal
from .annotation import Annotation


@_register_schema_internal
class Tracklet(Annotation):
    """A `Tracklet` is a temporal segment of a video sequence.

    Attributes:
        start_timestep: The start timestep of the tracklet.
        end_timestep: The end timestep of the tracklet.
        start_timestamp: The start timestamp of the tracklet.
        end_timestamp: The end timestamp of the tracklet.
    """

    start_timestep: int = -1
    end_timestep: int = -1
    start_timestamp: float = -1.0
    end_timestamp: float = -1.0

    @model_validator(mode="after")
    def _validate_fields(self) -> Self:
        if (
            self.start_timestep < -1
            or self.end_timestep < -1
            or self.start_timestamp < -1.0
            or self.end_timestamp < -1.0
        ):
            raise ValueError(
                "start_timestep, end_timestep, start_timestamp, and end_timestamp must be greater than or equal to -1."
            )
        elif (
            self.start_timestep == -1
            and self.end_timestep == -1
            and self.start_timestamp == -1.0
            and self.end_timestamp == -1.0
        ):
            raise ValueError(
                "At least one of start_timestep, end_timestep, start_timestamp, or end_timestamp must be set."
            )
        elif self.start_timestep == -1 and self.end_timestep > 0:
            raise ValueError("start_timestep must be set if end_timestep is set.")
        elif self.end_timestep == -1.0 and self.start_timestep > 0.0:
            raise ValueError("end_timestep must be set if start_timestep is set.")
        elif self.start_timestamp == -1.0 and self.end_timestamp > 0.0:
            raise ValueError("start_timestamp must be set if end_timestamp is set.")
        elif self.end_timestamp == -1.0 and self.start_timestamp > 0.0:
            raise ValueError("end_timestamp must be set if start_timestamp is set.")
        elif self.start_timestep > 0 and self.end_timestep > 0 and self.start_timestep > self.end_timestep:
            raise ValueError("start_timestep must be less than or equal to end_timestep.")
        elif self.start_timestamp > 0 and self.end_timestamp > 0 and self.start_timestamp > self.end_timestamp:
            raise ValueError("start_timestamp must be less than or equal to end_timestamp.")
        return self


def is_tracklet(cls: type, strict: bool = False) -> bool:
    """Check if the given class is a subclass of `Tracklet`."""
    return issubclass_strict(cls, Tracklet, strict)


def create_tracklet(
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    view_ref: ViewRef = ViewRef.none(),
    entity_ref: EntityRef = EntityRef.none(),
    source_ref: SourceRef = SourceRef.none(),
    start_timestep: int = -1,
    end_timestep: int = -1,
    start_timestamp: float = -1.0,
    end_timestamp: float = -1.0,
) -> Tracklet:
    """Create a `Tracklet` instance.

    Args:
        id: The tracklet id.
        item_ref: The item reference.
        view_ref: The view reference.
        entity_ref: The parent track reference.
        source_ref: The source reference.
        start_timestep: The start timestep of the tracklet.
        end_timestep: The end timestep of the tracklet.
        start_timestamp: The start timestamp of the tracklet.
        end_timestamp: The end timestamp of the tracklet.

    Returns:
        The created `Tracklet` instance.
    """
    return Tracklet(
        id=id,
        item_ref=item_ref,
        view_ref=view_ref,
        entity_ref=entity_ref,
        source_ref=source_ref,
        start_timestep=start_timestep,
        end_timestep=end_timestep,
        start_timestamp=start_timestamp,
        end_timestamp=end_timestamp,
    )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .dataset_builder import DatasetBuilder
from .folders import FolderBaseBuilder, ImageFolderBuilder, VideoFolderBuilder, VQAFolderBuilder


__all__ = ["DatasetBuilder", "FolderBaseBuilder", "ImageFolderBuilder", "VideoFolderBuilder", "VQAFolderBuilder"]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .pyarrow_utils import DESERIALIZE_PYARROW_DATATYPE, SERIALIZE_PYARROW_DATATYPE
from .schemas import (
    Annotation,
    BaseIntrinsics,
    BaseSchema,
    BBox,
    BBox3D,
    CamCalibration,
    Classification,
    CompressedRLE,
    Conversation,
    Embedding,
    Entity,
    Extrinsics,
    Image,
    Intrinsics,
    Item,
    KeyPoints,
    KeyPoints3D,
    Message,
    PointCloud,
    Relation,
    SchemaGroup,
    SequenceFrame,
    Source,
    SourceKind,
    Text,
    TextSpan,
    Track,
    Tracklet,
    Video,
    View,
    ViewEmbedding,
    create_bbox,
    create_bbox3d,
    create_cam_calibration,
    create_classification,
    create_compressed_rle,
    create_conversation,
    create_image,
    create_keypoints,
    create_keypoints3d,
    create_message,
    create_point_cloud,
    create_relation,
    create_sequence_frame,
    create_source,
    create_text,
    create_text_span,
    create_track,
    create_tracklet,
    create_video,
    create_view_embedding_function,
    group_to_str,
    is_annotation,
    is_base_schema,
    is_bbox,
    is_bbox3d,
    is_cam_calibration,
    is_classification,
    is_compressed_rle,
    is_conversation,
    is_embedding,
    is_entity,
    is_image,
    is_item,
    is_keypoints,
    is_keypoints3d,
    is_message,
    is_point_cloud,
    is_relation,
    is_sequence_frame,
    is_source,
    is_text,
    is_text_span,
    is_track,
    is_tracklet,
    is_video,
    is_view,
    is_view_embedding,
    register_schema,
    schema_to_group,
)
from .types import (
    AnnotationRef,
    BaseType,
    EmbeddingRef,
    EntityRef,
    ItemRef,
    NDArrayFloat,
    SchemaRef,
    SourceRef,
    ViewRef,
    create_annotation_ref,
    create_embedding_ref,
    create_entity_ref,
    create_item_ref,
    create_ndarray_float,
    create_schema_ref,
    create_source_ref,
    create_view_ref,
    is_annotation_ref,
    is_base_type,
    is_embedding_ref,
    is_entity_ref,
    is_item_ref,
    is_ndarray_float,
    is_schema_ref,
    is_source_ref,
    is_view_ref,
)


__all__ = [
    "DESERIALIZE_PYARROW_DATATYPE",
    "SERIALIZE_PYARROW_DATATYPE",
    "Annotation",
    "AnnotationRef",
    "BaseSchema",
    "BaseType",
    "BaseIntrinsics",
    "BBox",
    "BBox3D",
    "CamCalibration",
    "Classification",
    "CompressedRLE",
    "Conversation",
    "Embedding",
    "EmbeddingRef",
    "Entity",
    "EntityRef",
    "Extrinsics",
    "Image",
    "Intrinsics",
    "Item",
    "ItemRef",
    "KeyPoints",
    "KeyPoints3D",
    "Message",
    "NDArrayFloat",
    "PointCloud",
    "SchemaGroup",
    "SchemaRef",
    "SequenceFrame",
    "Source",
    "SourceKind",
    "SourceRef",
    "TextSpan",
    "Relation",
    "Text",
    "Track",
    "Tracklet",
    "Video",
    "View",
    "ViewEmbedding",
    "ViewRef",
    "create_annotation_ref",
    "create_bbox",
    "create_bbox3d",
    "create_cam_calibration",
    "create_classification",
    "create_compressed_rle",
    "create_conversation",
    "create_embedding_ref",
    "create_entity_ref",
    "create_image",
    "create_item_ref",
    "create_keypoints",
    "create_keypoints3d",
    "create_text_span",
    "create_message",
    "create_ndarray_float",
    "create_point_cloud",
    "create_relation",
    "create_schema_ref",
    "create_sequence_frame",
    "create_source",
    "create_source_ref",
    "create_text",
    "create_track",
    "create_tracklet",
    "create_video",
    "create_view_ref",
    "create_view_embedding_function",
    "group_to_str",
    "is_annotation",
    "is_annotation_ref",
    "is_base_schema",
    "is_base_type",
    "is_bbox",
    "is_bbox3d",
    "is_cam_calibration",
    "is_classification",
    "is_compressed_rle",
    "is_conversation",
    "is_embedding",
    "is_embedding_ref",
    "is_entity",
    "is_entity_ref",
    "is_image",
    "is_item",
    "is_item_ref",
    "is_keypoints",
    "is_keypoints3d",
    "is_text_span",
    "is_message",
    "is_ndarray_float",
    "is_point_cloud",
    "is_relation",
    "is_schema_ref",
    "is_sequence_frame",
    "is_source",
    "is_source_ref",
    "is_text",
    "is_track",
    "is_tracklet",
    "is_video",
    "is_view",
    "is_view_embedding",
    "is_view_ref",
    "register_schema",
    "schema_to_group",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================


from pathlib import Path
from typing import Any

from pydantic import BaseModel
from typing_extensions import Self

from pixano.app.models.dataset_info import DatasetInfoModel
from pixano.datasets import DatasetFeaturesValues, DatasetInfo, DatasetSchema
from pixano.datasets.dataset import Dataset


class DatasetModel(BaseModel):
    """The model of a dataset.

    Attributes:
        id: Dataset ID.
        path: Path to the dataset.
        previews_path: Path to the previews.
        media_dir: Path to the media directory.
        thumbnail: Path to the thumbnail.
        dataset_schema: The dataset schema.
        feature_values: The feature values of the dataset.
        info: The dataset info.
    """

    id: str
    path: Path
    previews_path: Path
    media_dir: Path
    thumbnail: Path
    dataset_schema: DatasetSchema
    feature_values: DatasetFeaturesValues
    info: DatasetInfoModel

    @classmethod
    def from_dataset(cls, dataset: Dataset) -> Self:
        """Create a dataset model from a dataset.

        Args:
            dataset: The dataset.

        Returns:
            The dataset model.
        """
        info = DatasetInfoModel.from_dataset_info(dataset.info, dataset.path)
        return cls(
            id=dataset.info.id,
            path=dataset.path,
            previews_path=dataset.previews_path,
            media_dir=dataset.media_dir,
            thumbnail=dataset.thumbnail,
            dataset_schema=dataset.schema,
            feature_values=dataset.features_values,
            info=info,
        )

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Self:
        """Create a dataset model from a JSON object.

        Args:
            data: JSON object.

        Returns:
            Dataset model.
        """
        return cls(
            id=data["id"],
            path=Path(data["path"]),
            previews_path=Path(data["previews_path"]),
            media_dir=Path(data["media_dir"]),
            thumbnail=Path(data["thumbnail"]),
            dataset_schema=DatasetSchema.deserialize(data["dataset_schema"]),
            feature_values=DatasetFeaturesValues(**data["feature_values"]),
            info=DatasetInfoModel.from_dataset_info(DatasetInfo(**data["info"]), Path(data["path"])),
        )

    def to_dataset(self) -> Dataset:
        """Create a dataset from a dataset model."""
        return Dataset(self.path, self.media_dir)


class PaginationColumn(BaseModel):
    """Column description.

    Attributes:
        name: column name.
        type: column type.
    """

    name: str
    type: str


class TableData(BaseModel):
    """Table data.

    Attributes:
        columns: column descriptions.
        rows: rows (actual data).
    """

    columns: list[PaginationColumn]
    rows: list[dict[str, Any]]
    # Note: Any is one of the allowed cell types (int, float, str, bool, Image/Video/..., graph, ...)


class PaginationInfo(BaseModel):
    """Pagination info.

    Attributes:
        current_page: current page.
        page_size: number of items per page.
        total_size: total number of items.
    """

    current_page: int
    page_size: int
    total_size: int


class DatasetBrowser(BaseModel):
    """Data for Dataset Browser page.

    Attributes:
        id: dataset id
        name: dataset name
        table_data: table data
        pagination: pagination infos
        semantic_search: list of semantic search available models
    """

    id: str
    name: str
    table_data: TableData
    pagination: PaginationInfo
    semantic_search: list[str]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================


import json
from typing import Any

from fastapi.encoders import jsonable_encoder

from pixano.features import Source

from ..dataset_info import DatasetInfo
from ..dataset_schema import DatasetItem
from .dataset_exporter import DatasetExporter


class DefaultJSONLDatasetExporter(DatasetExporter):
    """Default JSON Lines dataset exporter."""

    def initialize_export_data(self, info: DatasetInfo, sources: list[Source]) -> list[dict[str, Any]]:
        """Initialize a list of dictionaries to be exported.

        The first line contains the following elements:
        - dataset info
        - the sources

        Args:
            info: The dataset information.
            sources: The list of sources.

        Returns:
            A list of dictionaries containing the data to be exported.
        """
        export_data = [
            {"info": info.model_dump(), "sources": [s.model_dump(exclude_timestamps=True) for s in sources]}
        ]
        return export_data

    def export_dataset_item(
        self, export_data: list[dict[str, Any]], dataset_item: DatasetItem
    ) -> list[dict[str, Any]]:
        """Store the dataset item in the `export_data` list of dictionaries.

        Args:
            export_data: A list of dictionaries containing the dataset items to be exported.
            dataset_item: The dataset item to be exported.

        Returns:
            A list of dictionaries containing the dataset items to be exported.
        """
        export_data.append(dataset_item.model_dump(exclude_timestamps=True))
        return export_data

    def save_data(self, export_data: list[dict[str, Any]], split: str, file_name: str, file_num: int) -> None:
        """Save data to the specified directory.

        The saved directory has the following structure:
            export_dir/{split}_{file_name}_0.jsonl
                      /...
                      /{split}_{file_name}_{file_num}.jsonl
                      /...
                      /{split}_{file_name}_n.jsonl


        Args:
            export_data: The list of dictionaries containing the data to be saved.
            split: The split of the dataset item being saved.
            file_name: The name of the file to save the data in.
            file_num: The number of the file to save the data in.
        """
        info, data = export_data[0], export_data[1:]

        info_path = self.export_dir / "info.json"
        info_path.write_text(json.dumps(info), encoding="utf-8")

        json_path = self.export_dir / f"{split}_{file_name}_{file_num}.jsonl"
        json_path.write_text("\n".join([json.dumps(jsonable_encoder(d)) for d in data]), encoding="utf-8")



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================


class DatasetPaginationError(ValueError):
    """Error raised when paginating a dataset."""

    pass


class DatasetAccessError(ValueError):
    """Error raised when accessing a dataset."""

    pass


class DatasetWriteError(ValueError):
    """Error raised when writing to a dataset."""

    pass


class DatasetIntegrityError(ValueError):
    """Error raised when dataset integrity is compromised."""

    pass



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from datetime import datetime
from typing import Any, Generic, TypeVar

from pydantic import BaseModel, ConfigDict, field_validator
from typing_extensions import Self

from pixano.datasets import Dataset
from pixano.features import BaseSchema
from pixano.utils.validation import validate_and_init_create_at_and_update_at

from .table_info import TableInfo


T = TypeVar("T", bound=BaseSchema)


class BaseSchemaModel(BaseModel, Generic[T]):
    """Base schema model.

    This class is a base class for all schema models. It provides methods to convert a row to a model and vice
    versa.

    Attributes:
        id: Unique identifier of the row.
        created_at: The creation date of the row.
        updated_at: The last modification date of the row.
        table_info: Information about the table to which the row belongs.
        data: Dumped data from the Pixano backend row except the id.
    """

    model_config = ConfigDict(validate_assignment=True)
    id: str
    created_at: datetime
    updated_at: datetime
    table_info: TableInfo
    data: dict[str, Any]

    @field_validator("id", mode="after")
    @classmethod
    def _id_validator(cls, v: str) -> str:
        if " " in v:
            raise ValueError("id must not contain spaces")
        return v

    def __init__(self, /, created_at: datetime | None = None, updated_at: datetime | None = None, **data: Any):
        """Create a new model by parsing and validating input data from keyword arguments.

        Raises [`ValidationError`][pydantic_core.ValidationError] if the input data cannot be
        validated to form a valid model.

        `self` is explicitly positional-only to allow `self` as a field name.

        Args:
            created_at: The creation date of the object.
            updated_at: The last modification date of the object.
            data: The data of the object validated by Pydantic.
        """
        created_at, updated_at = validate_and_init_create_at_and_update_at(created_at, updated_at)
        data.update({"created_at": created_at, "updated_at": updated_at})
        super().__init__(**data)

    def model_dump(self, exclude_timestamps: bool = False, **kwargs: Any) -> dict[str, Any]:
        """Dump the model to a dictionary.

        Args:
            exclude_timestamps: Exclude timestamps "created_at" and "updated_at" from the model dump. Useful for
                comparing models without timestamps.
            kwargs: Arguments for pydantic `BaseModel.model_dump()`.

        Returns:
            The model dump.
        """
        model_dump = super().model_dump(**kwargs)
        if exclude_timestamps:
            model_dump.pop("created_at", None)
            model_dump.pop("updated_at", None)
        return model_dump

    @classmethod
    def from_row(cls, row: T, table_info: TableInfo) -> Self:
        """Create a model from a row.

        Args:
            row: The row to create the model from.
            table_info: The table info of the row.

        Returns:
            The created model.
        """
        model_dict = {}
        data = {}
        for key, value in row.model_dump().items():
            if key in ["id", "created_at", "updated_at"]:
                model_dict[key] = value
                continue
            data[key] = value
        model_dict["data"] = data
        model_dict["table_info"] = table_info

        return cls.model_validate(model_dict)

    @classmethod
    def from_rows(cls, rows: list[T], table_info: TableInfo) -> list[Self]:
        """Create a list of models from a list of schemas.

        Args:
            rows: The rows to create the models from.
            table_info: The table info of the rows.

        Returns:
            The list of created models.
        """
        return [cls.from_row(row, table_info) for row in rows]

    def to_row(self, dataset: Dataset) -> T:
        """Create a row from the model.

        Args:
            dataset: The dataset of the row.

        Returns:
            The created row.
        """
        schema_dict = self.model_dump()
        schema = dataset.schema.schemas[self.table_info.name]
        row = schema.model_validate(
            {
                "id": schema_dict["id"],
                "created_at": schema_dict["created_at"],
                "updated_at": schema_dict["updated_at"],
                **schema_dict["data"],
            }
        )
        row.dataset = dataset
        row.table_name = self.table_info.name
        return row

    @staticmethod
    def to_rows(models: list["BaseSchemaModel"], dataset: Dataset) -> list[T]:
        """Create a list of rows from a list of models.

        Args:
            models: The models to create the rows from.
            dataset: The dataset of the row.

        Returns:
            The list of created rows.
        """
        return [model.to_row(dataset) for model in models]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .annotations import (
    Annotation,
    BaseIntrinsics,
    BBox,
    BBox3D,
    CamCalibration,
    Classification,
    CompressedRLE,
    Extrinsics,
    Intrinsics,
    KeyPoints,
    KeyPoints3D,
    Message,
    Relation,
    TextSpan,
    Tracklet,
    create_bbox,
    create_bbox3d,
    create_cam_calibration,
    create_classification,
    create_compressed_rle,
    create_keypoints,
    create_keypoints3d,
    create_message,
    create_relation,
    create_text_span,
    create_tracklet,
    is_annotation,
    is_bbox,
    is_bbox3d,
    is_cam_calibration,
    is_classification,
    is_compressed_rle,
    is_keypoints,
    is_keypoints3d,
    is_message,
    is_relation,
    is_text_span,
    is_tracklet,
)
from .base_schema import BaseSchema, is_base_schema
from .embeddings import Embedding, ViewEmbedding, create_view_embedding_function, is_embedding, is_view_embedding
from .entities import (
    Conversation,
    Entity,
    Track,
    create_conversation,
    create_track,
    is_conversation,
    is_entity,
    is_track,
)
from .items import Item, is_item
from .registry import register_schema
from .schema_group import SchemaGroup, group_to_str, schema_to_group
from .source import Source, SourceKind, create_source, is_source
from .views import (
    Image,
    PointCloud,
    SequenceFrame,
    Text,
    Video,
    View,
    create_image,
    create_point_cloud,
    create_sequence_frame,
    create_text,
    create_video,
    is_image,
    is_point_cloud,
    is_sequence_frame,
    is_text,
    is_video,
    is_view,
)


__all__ = [
    "Annotation",
    "BaseIntrinsics",
    "BaseSchema",
    "BBox",
    "BBox3D",
    "CamCalibration",
    "Classification",
    "CompressedRLE",
    "Conversation",
    "Embedding",
    "Entity",
    "Extrinsics",
    "Image",
    "ImageObject",
    "Intrinsics",
    "Item",
    "KeyPoints",
    "KeyPoints3D",
    "Message",
    "TextSpan",
    "PointCloud",
    "Relation",
    "SchemaGroup",
    "SequenceFrame",
    "Source",
    "SourceKind",
    "Text",
    "Track",
    "Tracklet",
    "Video",
    "View",
    "ViewEmbedding",
    "create_bbox",
    "create_bbox3d",
    "create_cam_calibration",
    "create_classification",
    "create_compressed_rle",
    "create_conversation",
    "create_image",
    "create_image_object",
    "create_keypoints",
    "create_keypoints3d",
    "create_text_span",
    "create_message",
    "create_point_cloud",
    "create_relation",
    "create_sequence_frame",
    "create_source",
    "create_text",
    "create_track",
    "create_tracklet",
    "create_video",
    "create_view_embedding_function",
    "group_to_str",
    "is_annotation",
    "is_base_schema",
    "is_bbox",
    "is_bbox3d",
    "is_cam_calibration",
    "is_classification",
    "is_compressed_rle",
    "is_conversation",
    "is_embedding",
    "is_entity",
    "is_image",
    "is_item",
    "is_keypoints",
    "is_keypoints3d",
    "is_text_span",
    "is_message",
    "is_point_cloud",
    "is_relation",
    "is_sequence_frame",
    "is_source",
    "is_text",
    "is_track",
    "is_tracklet",
    "is_view",
    "is_video",
    "is_view_embedding",
    "register_schema",
    "schema_to_group",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pathlib import Path

from pixano.datasets.dataset_info import DatasetInfo
from pixano.datasets.dataset_schema import DatasetItem
from pixano.datasets.workspaces import WorkspaceType
from pixano.datasets.workspaces.dataset_items import DefaultVideoDatasetItem

from .base import FolderBaseBuilder


# TODO: Add more video extensions supported by ffmpeg
VIDEO_EXTENSIONS = [
    ".mp4",
    ".avi",
    ".mov",
    ".mkv",
    ".webm",
    ".flv",
    ".vob",
]


class VideoFolderBuilder(FolderBaseBuilder):
    """Builder for video datasets stored in a folder."""

    EXTENSIONS = VIDEO_EXTENSIONS
    WORKSPACE_TYPE = WorkspaceType.VIDEO

    def __init__(
        self,
        source_dir: Path | str,
        target_dir: Path | str,
        info: DatasetInfo,
        dataset_item: type[DatasetItem] = DefaultVideoDatasetItem,
        url_prefix: Path | str | None = None,
    ) -> None:
        """Initialize the `VideoFolderBuilder`.

        Args:
            source_dir: The source directory for the dataset.
            target_dir: The target directory for the dataset.
            dataset_item: The dataset item schema.
            info: User informations (name, description, ...) for the dataset.
            url_prefix: The path to build relative URLs for the views. Useful to build dataset libraries to pass the
                relative path from the media directory.
        """
        super().__init__(
            source_dir=source_dir, target_dir=target_dir, dataset_item=dataset_item, info=info, url_prefix=url_prefix
        )
